"""Entry point for generating gifs in "meme" format

Functions in the module are used in the process of
generating gifs in the format of white background
and black impact font given a pre-existing gif.

Functions are specific to gif creation, and must
be done in a specific order to get the desired
result.

Functions
---------
create_caption_background(lines, img) : Image
  Returns a new Image object with a white
  background with size corresponding to the amount of lines
get_text_wrap(text, font, max_width) : List[str]
  Returns a List of str, with the number of words per element
  being dependent on the rendered size of the font and the
  desired max width
create_caption(img):
  Returns a final Image with the caption
  Doesn't return an entire gif. If a gif is inputted, the
  function would only return the first frame of the gif.

"""
from PIL import Image, ImageDraw, ImageFont
from typing import List

def create_caption_background(
  lines: int, line_height: int, img: Image.Image
) -> Image.Image:
  """Returns a new Image object with a white
  background with size corresponding to the amount of lines

  Parameters
  ----------
  lines : int
    the number of lines to render on top of the white background
  line_height : int
    the maximum height that a line in a desired font would take
  img : Image
    the starting Image to format

  Returns
  -------
  Image
    a new Image with a white background on top
  """
  result = Image.new(
    'RGBA', (img.width, img.height + lines * line_height), (255, 255, 255)
  )
  result.paste(img, (0, lines * line_height + 20))
  return result

def get_text_wrap(
  text: str, font: ImageFont.FreeTypeFont, max_width: int
) -> List[str]:
  """Returns a List of str, with the number of words per element
  being dependent on the rendered size of the font and the
  desired max width

  Parameters
  ----------
  text : str
    a string of text to split up into lines
  font : FreeTypeFont
    a TrueType or OpenType font that determines words for each line
  max_width : int
    an int specifying the absolute max length, in pixels, that a line
    can take up

  Returns
  -------
  List[str]
    a List of strings separated by rendered size of the text
  """
  lines = []

  if font.getsize(text)[0] <= max_width:
    lines.append(text)
  else:
    words = text.split(' ')
    i = 0
    while i < len(words):
      line = ''
      while i < len(words) and font.getsize(line + words[i])[0] <= max_width:
        line = line + words[i] + ' '
        i += 1
        print(line)
      if not line:
        line = words[i]
        i += 1

      lines.append(line)
    return lines

def create_caption(
  caption: str, font: ImageFont.FreeTypeFont, img: Image.Image
) -> Image.Image:
  """Returns a final Image with the caption
  Doesn't return an entire gif. If a gif is inputted, the
  function would only return the first frame of the gif.

  Parameters
  ----------
  caption : str
    text to use as caption on top of image
  img : Image
    the starting Image to format

  Returns
  -------
  Image
    a new Image with a completed caption on top
  """
  text_lines = get_text_wrap(caption, font, img.width * 0.8)

  img = create_caption_background(len(text_lines), font.getsize('hg')[1], img)

  canvas = ImageDraw.Draw(img)
  image_width = img.width
  margin = 10
  for line in text_lines:
    font_width = canvas.textsize(line, font=font)[0]
    canvas.text(((image_width-font_width) / 2, margin), line,
     font=font, fill=(0,0,0))
    margin += font.getsize('hg')[1]
  del canvas

  return img



"""Classes used for management and storage of pre-defined
gif templates. Instances of these classes are to be
used in gif generation (inputted into generation
functions) and saved

Classes
-------
Gif
  An existing gif online
GifTemplate
  An existing gif online with methods for formatting
  that inherits from Gif
"""
from PIL import Image, ImageFont, ImageSequence
from io import BytesIO
from urllib.request import urlopen
from typing import List
from generator import create_caption

class Gif:
  """Basic representation of an online gif

  Fields
  ------
  name : str
    a user-given title for an online gif
  frames : List[Image]
    a List of Image objects that represent the frames
    of the gif

  """
  def __init__(self, name: str, frames: List[Image.Image]) -> None:
    """Creates a Gif object

    Parameters
    ----------
    name : str
      a user-given title of the template
    frames : List[Image]
      a List of Image objects taht represent the frames of the gif

    Methods
    -------
    GifTemplate : GifTemplate
      creates a GifTemplate object
    open_image : Image
      returns an Image instance corresponding to the link field
    """
    self.name = name
    self.frames = frames

  def save_gif(self) -> BytesIO:
    """Saves a GIF to BytesIO"""
    gif_bytes = BytesIO()
    self.frames[0].save(
      gif_bytes,
      format='GIF',
      save_all=True,
      append_images=self.frames[1:]
    )
    return gif_bytes
  def save_gif_fs(self, file_name) -> None:
    """Saves a GIF to the filesystem

    Parameters
    ----------
    file_name : str
      the name the file should be set as
    """
    self.frames[0].save(
      file_name,
      format='GIF',
      save_all=True,
      append_images=self.frames[1:]
    )

class GifTemplate(Gif):
  """An existing gif online for use in formatting

  Fields
  ------
  link : str
    a user-given image url for an online gif

  Methods
  -------
  GifTemplate : GifTemplate
    creates a new GifTemplate object
  generate_white_bg_black_impact(caption, font_size) : Image
    returns an Image (gif) with the format mentioned in the method name
  """
  def __init__(self, name: str, link: str):
    self.link = link
    super().__init__(
      name,
      [frame for frame in ImageSequence.Iterator(self.open_image())]
    )

  def open_image(self) -> Image.Image:
    """Returns an Image instance corresponding to the link field

    Returns
    -------
    Image
      an Image grabbed from this object's link field
    """
    return Image.open(urlopen(self.link))

  def generate_white_bg_black_impact(
    self, name: str, caption : str, font_size=42
  ) -> Gif:
    """Returns a List of Images (GIF) with the white background, black impact
    font format

    Parameters
    ----------
    name : str
      title for gif
    caption : str
      text to put on a white background in Impact font
    font_size : str
      size of the Impact font to be rendered onto the GIF

    Returns
    -------
    List[Image]
      a list of frames with each frame having been molded
      into the desired format
    """
    frames = []
    for frame in ImageSequence.Iterator(self.open_image()):
      frame = create_caption(
        caption,
        ImageFont.truetype('fonts/impact.ttf', font_size),
        self.open_image()
      )
      frames.append(frame)

    return Gif(name, frames)
